@@ -188,13 +188,12 @@ class notifier(object):
         mapfile = self.ui.config('notify', 'style')
         template = (self.ui.config('notify', hooktype) or
                     self.ui.config('notify', 'template'))
-        self.t = cmdutil.changeset_templater(self.ui, self.repo,
-                                             False, None, mapfile, False)
         if not mapfile and not template:
             template = deftemplates.get(hooktype) or single_template
         if template:
             template = templater.parsestring(template, quoted=False)
-            self.t.use_template(template)
+        self.t = cmdutil.changeset_templater(self.ui, self.repo, False, None,
+                                             template, mapfile, False)
 
     def strip(self, path):
         '''strip leading slashes from local path, turn into web-safe path.'''
@@ -341,8 +340,12 @@ class notifier(object):
 
         maxdiff = int(self.ui.config('notify', 'maxdiff', 300))
         prev = ctx.p1().node()
-        ref = ref and ref.node() or ctx.node()
-        chunks = patch.diff(self.repo, prev, ref, opts=patch.diffopts(self.ui))
+        if ref:
+            ref = ref.node()
+        else:
+            ref = ctx.node()
+        chunks = patch.diff(self.repo, prev, ref,
+                            opts=patch.diffallopts(self.ui))
         difflines = ''.join(chunks).splitlines()
 
         if self.ui.configbool('notify', 'diffstat', True):
